import numpy as np
cimport numpy as np

from pyrossgeo.__defs__ cimport node, cnode, transporter, model_term, DTYPE_t
from pyrossgeo.__defs__ import DTYPE, contact_scaling_types

#ctypedef void (*SIM_EVENT)(Simulation cg, int step_i, DTYPE_t t, DTYPE_t dt, DTYPE_t[:] X_state, DTYPE_t[:] dX_state)
#cdef void SIM_EVENT_NULL(Simulation cg, int step_i, DTYPE_t t, DTYPE_t dt, DTYPE_t[:] X_state, DTYPE_t[:] dX_state)

cdef class Simulation:
    cdef readonly int age_groups
    cdef readonly int model_dim
    cdef readonly int max_node_index

    cdef node* nodes
    cdef int nodes_num

    cdef cnode* cnodes
    cdef int cnodes_num

    cdef int state_size
    cdef int node_states_len # The slice of X_states which only comprises nodes

    cdef readonly object node_mappings
    cdef readonly object cnode_mappings

    # Model
    cdef model_term* model_linear_terms
    cdef int model_linear_terms_len
    cdef model_term* model_infection_terms
    cdef int model_infection_terms_len
    cdef np.ndarray infection_classes_indices
    cdef int infection_classes_num
    cdef object _Is_arr
    cdef np.ndarray _Is
    cdef object _Ns_arr
    cdef np.ndarray _Ns

    cdef np.ndarray contact_matrices
    cdef int** contact_matrices_at_each_loc # Contains what contact matrices are used at each location
    cdef int* contact_matrices_at_each_loc_len
    cdef int** contact_matrices_at_each_to # Contains what contact matrices are used for each destination (cnodes)
    cdef int* contact_matrices_at_each_to_len
    cdef readonly dict contact_matrices_key_to_index

    # Transport
    cdef transporter* Ts # Going into commuterverses
    cdef int Ts_num
    cdef transporter* cTs # Going out from commuterverses
    cdef int cTs_num

    # Used for lambda calculation
    cdef int*** nodes_at_j
    cdef int** nodes_at_j_len

    # Used for tau calculation
    cdef int*** cnodes_into_k
    cdef int** cnodes_into_k_len

    # Transport profile
    cdef DTYPE_t transport_profile_integrated
    cdef DTYPE_t transport_profile_integrated_r
    cdef DTYPE_t transport_profile_m
    cdef DTYPE_t transport_profile_c
    cdef DTYPE_t transport_profile_c_r

    # Stochasticity
    cdef readonly bint stochastic_simulation
    cdef np.ndarray stochastic_threshold_from_below # If all classes go above their threshold, start deterministic
    cdef np.ndarray stochastic_threshold_from_above # If any class go below their threshold, start stochastic

    # Infection scaling
    cdef int contact_scaling_type
    cdef np.ndarray contact_scaling_params
    cdef readonly np.ndarray location_area
    cdef readonly np.ndarray commuterverse_area

    # Events
    cdef list event_functions
    cdef list event_times
    cdef list event_repeat_times

    # Misc
    cdef readonly dict storage # Persistent storage that will be used for events
    cdef readonly object has_been_initialized # Python bool. If True, then the simulation has been initialized. 
    cdef public bint measure_sim_times

    #cdef csimulate(self, DTYPE_t[:] X_state, DTYPE_t t_start, DTYPE_t t_end, object _dts, int steps_per_save=*,
    #                            str out_file=*, int steps_per_print=*, bint only_save_nodes=*,
    #                            int steps_per_event=*,object event_function=*,
    #                            int steps_per_cevent=*, SIM_EVENT cevent_function=*)

    cpdef get_contact_matrix_keys(self)
    cpdef get_contact_matrix(self, str cmat_key)
    cpdef set_contact_matrix(self, str cmat_key, np.ndarray cmat)

    cpdef stop_commuting(self, bint s)
    cpdef bint is_commuting_stopped(self)




import numpy as np
import pickle
import zarr

def extract_node_data(sim_data):
    """Returns the results of the simulation for each node.

    Parameters
    ----------
        sim_data : Tuple
            Simulation data.

    Returns
    -------
    dict
        A dictionary of keys of the form `(i, j)`, corresponding to
        home node, and location node respectively.
        `node_data[i,j,k]` is an `np.ndarray` of shape
        `(ts.size, &#35; of age groups, &#35; of classes)`.
    """
    node_mappings, cnode_mappings, ts, X_states = sim_data
    node_data = {}

    age_groups = 0
    model_dim = 0
    for a,o,i,j in node_mappings:
        if a+1 > age_groups:
            age_groups = a+1
        if o+1 > model_dim:
            model_dim = o+1
    for a,o,i,j,k in cnode_mappings:
        if a+1 > age_groups:
            age_groups = a+1
        if o+1 > model_dim:
            model_dim = o+1

    for a,o,i,j in node_mappings:
        index = node_mappings[a,o,i,j]

        if not (i,j) in node_data:
            node_data[i,j] = np.zeros( (len(ts), age_groups, model_dim) )

        node_data[i,j][:,a,o] = X_states[:,index]

    return node_data

def extract_cnode_data(sim_data):
    """Returns the results of the simulation for each commuter node.

    Parameters
    ----------
        sim_data : Tuple
            Simulation data.

    Returns
    -------
    dict
        A dictionary of keys of the form `(i, j, k)`, corresponding to
        home node, origin node and destination node respectively.
        `cnode_data[i,j,k]` is an `np.ndarray` of shape
        `(ts.size, &#35; of age groups, &#35; of classes)`.
    """
    node_mappings, cnode_mappings, ts, X_states = sim_data
    cnode_data = {}

    age_groups = 0
    model_dim = 0
    for a,o,i,j in node_mappings:
        if a+1 > age_groups:
            age_groups = a+1
        if o+1 > model_dim:
            model_dim = o+1
    for a,o,i,j,k in cnode_mappings:
        if a+1 > age_groups:
            age_groups = a+1
        if o+1 > model_dim:
            model_dim = o+1

    for a,o,i,j,k in cnode_mappings:
        index = cnode_mappings[a,o,i,j,k]

        if not (i,j,k) in cnode_data:
            cnode_data[i,j,k] = np.zeros( (len(ts), age_groups, model_dim) )

        cnode_data[i,j,k][:,a,o] = X_states[:,index]

    return cnode_data

def extract_network_data(sim_data):
    """Returns the results of the simulation for the whole network.

    Parameters
    ----------
        sim_data : Tuple
            Simulation data.

    Returns
    -------
    np.ndarray
        An array of shape (ts.size, &#35; of age groups, &#35; of classes).
        It contains the result of the simulation of the network as a whole
        for each age group and class.
    """
    node_mappings, cnode_mappings, ts, X_states = sim_data

    age_groups = 0
    model_dim = 0
    for a,o,i,j in node_mappings:
        if a+1 > age_groups:
            age_groups = a+1
        if o+1 > model_dim:
            model_dim = o+1
    for a,o,i,j,k in cnode_mappings:
        if a+1 > age_groups:
            age_groups = a+1
        if o+1 > model_dim:
            model_dim = o+1

    network_data = np.zeros( (len(ts), age_groups, model_dim) )

    for a,o,i,j in node_mappings:
        network_data[:,a,o] += X_states[:,node_mappings[a,o,i,j]]

    for a,o,i,j,k in cnode_mappings:
        network_data[:,a,o] += X_states[:,cnode_mappings[a,o,i,j,k]]

    return network_data

def extract_location_data(sim_data):
    """Returns the results of the simulation for a given location.

    Parameters
    ----------
        sim_data : Tuple
            Simulation data.

    Returns
    -------
    np.ndarray
        An array of shape (ts.size, &#35; of age groups, &#35; of classes,
        &#35; of locations). It contains the results of the simulation at each
        location. So `community_data[5,0,1,32]` contains the state of
        people of age-bracket 0, class 1 who are at location 32, at step 5
        of the simulation.
    """
    node_mappings, cnode_mappings, ts, X_states = sim_data

    age_groups = 0
    model_dim = 0
    max_loc_index = 0

    for a,o,i,j in node_mappings:
        if a+1 > age_groups:
            age_groups = a+1
        if o+1 > model_dim:
            model_dim = o+1
        if j > max_loc_index:
            max_loc_index = j
    for a,o,i,j,_ in cnode_mappings:
        if a+1 > age_groups:
            age_groups = a+1
        if o+1 > model_dim:
            model_dim = o+1
        if j > max_loc_index:
            max_loc_index = j

    location_data = np.zeros( (len(ts), age_groups, model_dim, max_loc_index+1) )

    for a,o,i,j in node_mappings:
        location_data[:,a,o,j] += X_states[:,node_mappings[a,o,i,j]]

    return location_data

def extract_community_data(sim_data):
    """Returns the results of the simulation for each community.

    Parameters
    ----------
        sim_data : Tuple
            Simulation data.

    Returns
    -------
    np.ndarray
        An array of shape (ts.size, &#35; of age groups, &#35; of classes,
        &#35; of locations). It contains the results of the simulation summed
        over each community. So `community_data[:,0,1,32]` contains the
        history of all people of age-bracket 0, class 1 and who live at location 32.
    """
    node_mappings, cnode_mappings, ts, X_states = sim_data

    age_groups = 0
    model_dim = 0
    max_home_index = 0

    for a,o,i,j in node_mappings:
        if a+1 > age_groups:
            age_groups = a+1
        if o+1 > model_dim:
            model_dim = o+1
        if i > max_home_index:
            max_home_index = i
    for a,o,i,j,k in cnode_mappings:
        if a+1 > age_groups:
            age_groups = a+1
        if o+1 > model_dim:
            model_dim = o+1
        if i > max_home_index:
            max_home_index = i

    community_data = np.zeros( (len(ts), age_groups, model_dim, max_home_index+1) )

    for a,o,i,j in node_mappings:
        community_data[:,a,o,i] += X_states[:,node_mappings[a,o,i,j]]

    for a,o,i,j,k in cnode_mappings:
        community_data[:,a,o,i] += X_states[:,cnode_mappings[a,o,i,j,k]]

    return community_data

def extract_simulation_data(sim_data):
    """Returns a tuple containing various formatted data for a given simulation result.

    It returns `node_data, cnode_data, location_data, community_data, network_data`.
    """
    node_mappings, cnode_mappings, ts, _ = sim_data

    node_data = extract_node_data(sim_data)
    cnode_data = extract_cnode_data(sim_data)

    age_groups = 0
    model_dim = 0
    max_home_index = 0
    max_loc_index = 0

    for a,o,i,j in node_mappings:
        if a+1 > age_groups:
            age_groups = a+1
        if o+1 > model_dim:
            model_dim = o+1
        if i > max_home_index:
            max_home_index = i
        if j > max_loc_index:
            max_loc_index = j
    for a,o,i,j,k in cnode_mappings:
        if a+1 > age_groups:
            age_groups = a+1
        if o+1 > model_dim:
            model_dim = o+1
        if i > max_home_index:
            max_home_index = i
        if j > max_loc_index:
            max_loc_index = j

    location_data = np.zeros( (len(ts), age_groups, model_dim, max_loc_index+1) )
    community_data = np.zeros( (len(ts), age_groups, model_dim, max_home_index+1) )
    network_data = np.zeros( (len(ts), age_groups, model_dim) )

    for i,j in node_data:
        node_data_ij = node_data[i,j]
        location_data[:, :, :, j] += node_data_ij
        community_data[:, :, :, i] += node_data_ij
        network_data[:, :, :] += node_data_ij

    for i,j,k in cnode_data:
        cnode_data_ijk = cnode_data[i,j,k]
        community_data[:, :, :, i] += cnode_data_ijk
        network_data[:, :, :] += cnode_data_ijk

    return ts, node_data, cnode_data, location_data, community_data, network_data

def extract_ts(sim_data):
    """Returns the results of the simulation times given simulation data.

    Parameters
    ----------
        sim_data : Tuple
            Simulation data.

    Returns
    -------
    np.ndarray
        A 1D array containing each time-step.
    """
    ts_saved = sim_data[2]
    return ts_saved

def load_sim_data(load_path, use_zarr=False):
    """Loads 

    Parameters
    ----------
        load_path : str
            Path of the simulation data folder.
        use_zarr : bool, optional
            If True, the simulation data will be given as a zarr array,
            rather than as a numpy array. The former is useful if the
            data is very large.

    Returns
    -------
    tuple
        A tuple `(node_mappings, cnode_mappings, ts, X_states)`, containing
        all simulation data. `X_states` is either an `np.ndarray` or a `zarr.core.Array`.
        If `use_zarr=True`, the latter will be given.
    """
    node_mappings_path = 'node_mappings.pkl'
    cnode_mappings_path = 'cnode_mappings.pkl'
    ts_path = 'ts.npy'
    X_states_path = 'X_states.zarr'

    node_mappings = pickle.load( open( "%s/%s" %( load_path, node_mappings_path ), "rb" ) )
    cnode_mappings = pickle.load( open( "%s/%s" %( load_path, cnode_mappings_path ), "rb" ) )
    ts = np.load("%s/%s" %( load_path, ts_path ))
    X_states = zarr.open( "%s/%s" %( load_path, X_states_path ) , chunks=(len(ts), 1))

    if not use_zarr:
        X_states = X_states[:]

    sim_data = ( node_mappings, cnode_mappings, ts, X_states )
    return sim_data

def get_dt_schedule(times, end_time):
    """Generates a time-step schedule.

    Example:

    The following generates a time-step schedule where we use a time-step
    of one minute between 7-10 and 17-19 o\'clock, and 2 hours for all
    other times.

        ts, dts = pyrossgeo.utils.get_dt_schedule([
            (0,  2*60),
            (7*60,  1),
            (10*60, 2*60),
            (17*60, 1),
            (19*60, 2*60)
            ], end_time=24*60)

    Parameters
    ----------
        times : list
            list of tuples
        end_time : float
            The final time of the schedule.

    Returns
    -------
        tuple
            A tuple `(ts, dts)`. `dts` are the time-steps and `ts`
            the times.
    """
    times = list(times)
    times.append( (end_time, 0) )
    ts = []

    for i in range(len(times)-1):
        t, dt = times[i]
        t_next = times[i+1][0]
        ts.append(np.arange(t, t_next, dt))
        
    ts.append([end_time])
    ts = np.concatenate(ts)
    dts = (ts - np.roll(ts, 1))[1:]
        
    return np.array(ts, dtype=np.double), np.array(dts, dtype=np.double)

def correct_contact_matrix(cmat, Ni_old, Ni):
    """Corrects a contact matrix so that it satisfies consistency relations.
    """

def construct_local_contact_matrices(cmat, loc_pops, scaling_function):
    """Constructs local contact matrices from a global contact matrix.

    Constructs local contact matrices from a global contact matrix that
    scale with population density as well as obey local 
    """

def construct_spatial_contact_matrix(density_scaling_b, work_ratio, age_groups, node_populations, areas, commuter_network):
    """A wrapper for `pyross.contactMatrix.SpatialContactMatrix`.

    Uses PyRossGeo input files to construct a spatial contact matrix.

    Parameters
    ----------
        density_scaling_b : float
            Density scaling parameter.
        work_ratio : float
            Work ratio.
        age_groups : int
            The number of age groups.
        node_populations : string or pandas.DataFrame
            Path to the node population configuration file, or the configuration file itself.
        areas : np.array
            1D array, containing the area of each location
        commuter_network : string or pandas.DataFrame
            Path to the commuter network configuration file, or the configuration file itself.
            The helper function assumes that commutes are cyclical of length 2. In other
            words, all commuting edges come in pairs: One commuting edge
            leading from the home node to the destination node, and one
            leading back to the home node.

    Returns
    -------
        pyross.contactMatrix.SpatialContactMatrix
            The spatial contact matrix.
    """
    
    import pandas as pd
    import pyross

    if type(node_populations) == str:
        node_populations = pd.DataFrame(node_populations)
    if type(commuter_network) == str:
        commuter_network = pd.DataFrame(commuter_network)

    max_node_index = max( commuter_network.iloc[:,0:3].max().max(), node_populations.iloc[:,0:2].max().max() )

    compartment_pops = node_populations.iloc[:,2:]
    populations = np.zeros( (max_node_index+1, age_groups) )
    
    for i in range(age_groups):
        populations[:,i] = compartment_pops.iloc[:,i*age_groups:(i+1)*age_groups].sum(axis=1)

    commutes = np.zeros( (max_node_index+1, max_node_index+1, age_groups) )
    sel = commuter_network[ commuter_network.iloc[:,0] == commuter_network.iloc[:,1] ]
    commutes[ sel.iloc[:,1], sel.iloc[:,2], sel.iloc[:,3] ] = sel.iloc[:,4]

    scm_generator = pyross.contactMatrix.SpatialContactMatrix(density_scaling_b, work_ratio, populations, areas, commutes)
    return populations, commutes, scm_generator



import numpy as np
cimport numpy as np

from pyrossgeo._initialization import initialize
from pyrossgeo._simulation cimport simulate
from pyrossgeo._helpers import compute, free_sim

#cdef void SIM_EVENT_NULL(Simulation sim, int step_i, DTYPE_t t, DTYPE_t dt, DTYPE_t[:] X_state, DTYPE_t[:] dX_state):
#    raise(Exception('SIM_EVENT_NULL triggered'))

cdef class Simulation:
    """Simulates the geographic compartmental model.

    Attributes
    ----------
    age_groups : int
        The number of age-groups.
    model_dim : int
        The number of model classes.
    max_node_index : int
        The highest index of all nodes (not including cnodes). "Index" here is not to be confused
        with `(age, model_class, home, location)`.
    node_mappings : dict
        Contains mappings between nodes `(age, model_class, home, location)`
        to its index in the state vector `X_state`.
    cnode_mappings : dict
        Contains mappings between cnodes `(age, model_class, home, from, to)`
        to its index in the state vector `X_state`.
    stochastic_simulation : bool
        If True, simulation will be stochastic.
    storage : dict
        Persistent storage used for events.
    has_been_initialized : bool
        If True, then Simulation.initialize has been called.

    """

    def __cinit__(self):
        self.storage = {}
        self.has_been_initialized = False

        self.event_functions = []
        self.event_times = []
        self.event_repeat_times = []

    def __dealloc__(self):
        free_sim(self)

    def initialize(self, sim_config_path='', model_dat='', commuter_networks_dat='',
                        node_parameters_dat='', cnode_parameters_dat='',
                        contact_matrices_dat='', node_cmatrices_dat='',
                        cnode_cmatrices_dat='', node_populations_dat='',
                        cnode_populations_dat=''):
        """Initializes the simulation using the given configuration files.
        
        Each argument will be accepted as either as raw data or a path
        to a file containing the data. See the documentation for the
        format each configuration file should take.

        Parameters
        ----------
            model_dat : str or dict
                Specifies the epidemic model
            commuter_networks_dat : str or np.ndarray
                Specifies the commuter network
            node_parameters_dat : str or pandas.DataFrame
                Specifies the parameters of the model at each node
            cnode_parameters_dat : str or pandas.DataFrame
                Specifies the parameters of the model at each commuter node
            contact_matrices_dat : str or dict
                Specifies the contact matrices
            node_cmatrices_dat : str or pandas.DataFrame
                Specifies what contact matrix to use at each node
            cnode_cmatrices_dat : str or pandas.DataFrame
                Specifies what contact matrix to use at each commuter node
            node_populations_dat : str or np.ndarray
                Specifies the population at each node
            cnode_populations_dat : str or np.ndarray
                Specifies the population at each commuter node (default None)
        """
        return initialize(self, sim_config_path, model_dat, commuter_networks_dat,
                        node_parameters_dat, cnode_parameters_dat,
                        contact_matrices_dat, node_cmatrices_dat,
                        cnode_cmatrices_dat, node_populations_dat,
                        cnode_populations_dat)

    def simulate(self, X_state, t_start, t_end, dts, steps_per_save=-1,
                    save_path="", only_save_nodes=False, steps_per_print=-1):
        """Simulates the system.
        
        Simulates the system between times `t_start` and `t_end`, with the 
        initial condition `X_state`. `dts` an array of time-steps, where
        `dts[i]` is the time-step used during step `i` of the simualtion.

        If `save_path` is specified, then the result of the simulation 
        will be outputted directly to the hard-drive as a `zarr` array
        at the given path.

        Parameters
        ----------
            X_state : np.ndarray
                Initial condition of the system.
            t_start : float
                Start time
            t_end : float
                End time
            dts : float, or list or array of floats
                Time steps
            steps_per_save : int
                Number of simulation steps per saving the state (default -1)
            save_path : str
                The path of the folder to save the output to (default "")
            only_save_nodes : bool
                If True, commuter nodes will not be saved (default False)

        Returns
        -------
            tuple
                A tuple `((node_mappings, cnode_mappings), ts, X_states)`.
                `X_states` is an np.ndarray of shape `(ts.size, N)` where `N` is the total
                degrees of freedom of the system. If `only_save_nodes = True`
                then `N` is just the degrees of freedom of the nodes, and excludes
                the commuting nodes. `ts` is the times corresponding to `X_states`.
                `node_mappings` is a dictionary with keys of form `(a,o,i,j)`,
                corresponding to age-bracket, class, home and location respectively.
                `node_mappings[a,o,i,j]` is the column of `X_states` for the 
                corresponding state value. Similarly, `cnode_mappings` is
                a dictionary with keys of the form `(a,o,i,j,k)`, corresponding
                to age-bracket, class, home, origin, destination respectively.
        """
        return simulate(self, X_state, t_start, t_end, dts, steps_per_save,
                    save_path, only_save_nodes, steps_per_print)

    def compute(self, X_state, dX_state, t, dt):
        """Computes the right-hand side of the dynamical system.
        
        Sets the array `dX_state` to the derivative of the dynamical system
        at the state `X_state`.

        Parameters
        ----------
            X_state : np.ndarray
                The state of the system.
            dX_state : np.ndarray
                The array to input the derivative into.
            t : float
                Time
            dt : float
                The time-step used
        """
        compute(self, X_state, dX_state, t, dt)

    cpdef get_contact_matrix_keys(self):
        """Returns a list of the contact matrix keys."""
        return list(self.contact_matrices_key_to_index.keys())

    cpdef get_contact_matrix(self, str cmat_key):
        """Returns the contact matrix with the given key."""
        return self.contact_matrices[self.contact_matrices_key_to_index[cmat_key]]

    cpdef set_contact_matrix(self, str cmat_key, np.ndarray cmat):
        """Change the contact matrix of the given key.

        Parameters
        ----------
            cmat_key: the key of the contact matrix
            cmat: the array to set the contact matrix to.
        """
        self.contact_matrices[self.contact_matrices_key_to_index[cmat_key]] = cmat

    cpdef stop_commuting(self, bint s):
        """Disables the commuter network if s is True, and enables it if False."""

        # Turn off commuting
        if not self.is_commuting_stopped() and s:
            if not 'turn_off_Ts' in self.storage:
                self.storage['turn_off_Ts'] = []
            if not 'turn_off_cTs' in self.storage:
                self.storage['turn_off_cTs'] = []

            turn_off_Ts = self.storage['turn_off_Ts']
            turn_off_cTs = self.storage['turn_off_cTs']

            # Turn off commuting by setting t2=-1 for all transporters,
            # and store the t2 values so that commuting can be turned on again

            for i in range(self.Ts_num):
                turn_off_Ts.append( self.Ts[i].t2 )
                self.Ts[i].t2 = -1

            for i in range(self.cTs_num):
                turn_off_cTs.append( self.cTs[i].t2 )
                self.cTs[i].t2 = -1

            self.storage['commuting_is_stopped'] = True
        # Turn on commuting
        elif self.is_commuting_stopped():
            turn_off_Ts = self.storage['turn_off_Ts']
            turn_off_cTs = self.storage['turn_off_cTs']

            # Restore the t2 values of the transporters

            for i, t2 in zip(range(self.Ts_num), turn_off_Ts):
                self.Ts[i].t2 = t2

            for i, t2 in zip(range(self.cTs_num), turn_off_cTs):
                self.cTs[i].t2 = t2

            self.storage['commuting_is_stopped'] = False

    cpdef bint is_commuting_stopped(self):
        """Returns True if the commuter network has been disabled."""
        if 'commuting_is_stopped' in self.storage:
            return self.storage['commuting_is_stopped']
        else:
            return False

    def add_event(self, event_times, event_function, repeat_time=-1):
        """Adds an event function to be called at specific times of the simulation.

        Parameters
        ----------
            event_times : list or np.ndarray
                The times at which the event function is to be called.
            event_function : function
                The event function.
            repeat_time : float, optional
                If specified, then the event will be triggered every
                time `event_time == simulation_time % repeat_time`,
                for each time in `event_times`.
        """
        self.event_times.append(event_times)
        self.event_functions.append(event_function)
        self.event_repeat_times.append(repeat_time)

    #TODO
    # - function that returns what contact matrix is sued for a specific node
    # - see what the parameters are at a specific node (in dictionary format), and edit them
    # - lookup properties of commuter network


# distutils: language = c++
# distutils: extra_compile_args = -std=c++11

import numpy as np
cimport numpy as np

from pyrossgeo.__defs__ cimport node, cnode, transporter, model_term, DTYPE_t
from pyrossgeo.__defs__ import DTYPE
from pyrossgeo.Simulation cimport Simulation

cdef simulate(Simulation self, DTYPE_t[:] X_state, DTYPE_t t_start, DTYPE_t t_end, object _dts, int steps_per_save=*,
                            str save_path=*, bint only_save_nodes=*, int steps_per_print=*,
                            int random_seed=*)


cdef extern from "<random>" namespace "std":
    cdef cppclass mt19937:
        mt19937() # we need to define this constructor to stack allocate classes in Cython
        mt19937(unsigned int seed) # not worrying about matching the exact int type for seed

    cdef cppclass poisson_distribution[T]:
        poisson_distribution()
        poisson_distribution(DTYPE_t a)
        T operator()(mt19937 gen) # ignore the possibility of using other classes for "gen"


from libc.stdlib cimport free

from pyrossgeo.__defs__ cimport node, cnode, transporter, model_term, DTYPE_t
from pyrossgeo.__defs__ import DTYPE
from pyrossgeo.Simulation cimport Simulation

import numpy as np
cimport numpy as np

cpdef compute(Simulation self, DTYPE_t[:] X_state, DTYPE_t[:] dX_state, DTYPE_t t, DTYPE_t dt):
    t_start = t
    dts = np.array(dt)
    steps = 1
    steps_per_save = -1
    out_file = ''
    print_per_n_steps = -1
    
    self._simulate(self, X_state, dX_state, t_start, steps, dts, steps_per_save, out_file, print_per_n_steps)

cpdef free_sim(Simulation self):
    for ni in range(self.nodes_num):
        free(self.nodes[ni].incoming_T_indices)
        free(self.nodes[ni].outgoing_T_indices)
        free(self.nodes[ni].contact_matrix_indices)
        free(self.nodes[ni].linear_coeffs)
        free(self.nodes[ni].infection_coeffs)
    free(self.nodes)

    for ni in range(self.cnodes_num):
        free(self.cnodes[ni].contact_matrix_indices)
        free(self.cnodes[ni].linear_coeffs)
        free(self.cnodes[ni].infection_coeffs)
    free(self.cnodes)

    for ti in range(self.Ts_num):
        free(self.Ts[ti].moving_classes)
    free(self.Ts)

    for cti in range(self.cTs_num):
        free(self.cTs[cti].moving_classes)
    free(self.cTs)

    for age in range(self.age_groups):
        free(self.nodes_at_j_len[age])
        for loc in range(self.max_node_index+1):
            free(self.nodes_at_j[age][loc])
        free(self.nodes_at_j[age])
    free(self.nodes_at_j)
    free(self.nodes_at_j_len)

    for age in range(self.age_groups):
        free(self.cnodes_into_k_len[age])
        for to_k in range(self.max_node_index+1):
            free(self.cnodes_into_k[age][to_k])
        free(self.cnodes_into_k[age])
    free(self.cnodes_into_k)
    free(self.cnodes_into_k_len)

    for loc in range(self.max_node_index+1):
        free(self.contact_matrices_at_each_loc[loc])
    free(self.contact_matrices_at_each_loc)
    free(self.contact_matrices_at_each_loc_len)

    for to in range(self.max_node_index+1):
        free(self.contact_matrices_at_each_to[to])
    free(self.contact_matrices_at_each_to)
    free(self.contact_matrices_at_each_to_len)

    free(self.model_linear_terms)
    free(self.model_infection_terms)
    


import zarr
import pickle
import cython
cimport libc.math
from libc.stdlib cimport malloc, free
from cython.parallel import prange
import numpy as np
cimport numpy as np
import time # For seeding random
import scipy.stats

from pyrossgeo.__defs__ import DTYPE, contact_scaling_types

#@cython.wraparound(False)
#@cython.boundscheck(False)
#@cython.cdivision(True)
#@cython.nonecheck(False)
@cython.wraparound(False)
@cython.boundscheck(True)
@cython.cdivision(False)
@cython.nonecheck(True)
cdef simulate(Simulation self, DTYPE_t[:] X_state, DTYPE_t t_start, DTYPE_t t_end, object _dts, int steps_per_save=-1,
                            str save_path="", bint only_save_nodes=False, int steps_per_print=-1,
                            int random_seed=-1):

    if not self.has_been_initialized:
        raise Exception("Must initialise before starting simulation.")
    
    ####################################################################
    #### Definitions ###################################################
    ####################################################################

    #### Forward-Euler variables #######################################
    
    cdef DTYPE_t t = t_start
    cdef DTYPE_t dt
    cdef DTYPE_t r_dt
    cdef DTYPE_t[:] dts
    cdef DTYPE_t tday
    cdef int dts_num
    cdef int step_i
    cdef int steps
    cdef np.ndarray X_state_arr = np.asarray(X_state)

    #### System variables ##############################################

    cdef int age_groups = self.age_groups
    cdef int model_dim = self.model_dim
    cdef int max_node_index = self.max_node_index
    cdef int X_state_size = len(X_state)
    cdef node* nodes = self.nodes
    cdef int nodes_num = self.nodes_num
    cdef cnode* cnodes = self.cnodes
    cdef int cnodes_num = self.cnodes_num
    cdef int state_size = self.state_size
    cdef int node_states_len = self.node_states_len

    cdef np.ndarray total_N_arr = np.zeros( age_groups )
    cdef np.ndarray Ns_arr = np.zeros( (max_node_index+1, age_groups) )
    cdef np.ndarray cNs_arr = np.zeros( (max_node_index+1, age_groups) )
    cdef np.ndarray Os_arr = np.zeros( (max_node_index+1, age_groups, model_dim) )
    cdef np.ndarray cOs_arr = np.zeros( (max_node_index+1, age_groups, model_dim) )
    
    cdef DTYPE_t[:] total_N = total_N_arr
    cdef DTYPE_t[:,:] Ns = Ns_arr
    cdef DTYPE_t[:,:] cNs = cNs_arr
    cdef DTYPE_t[:,:,:] Os = Os_arr
    cdef DTYPE_t[:,:,:] cOs = cOs_arr

    # Measuring simulation times
    cdef bint measure_sim_times = self.measure_sim_times
    cdef DTYPE_t t_contact, t_node, t_cnode, t_T, t_cT, t_ff, t_rest

    # Model
    cdef model_term* model_linear_terms = self.model_linear_terms
    cdef int model_linear_terms_len = self.model_linear_terms_len
    cdef model_term* model_infection_terms = self.model_infection_terms
    cdef int model_infection_terms_len = self.model_infection_terms_len
    cdef int[:] infection_classes_indices = self.infection_classes_indices
    cdef int infection_classes_num = self.infection_classes_num

    cdef DTYPE_t[:,:,:] contact_matrices = self.contact_matrices
    cdef int contact_matrices_num = self.contact_matrices.shape[0]
    cdef int** contact_matrices_at_each_loc = self.contact_matrices_at_each_loc
    cdef int* contact_matrices_at_each_loc_len = self.contact_matrices_at_each_loc_len
    cdef int** contact_matrices_at_each_to = self.contact_matrices_at_each_to
    cdef int* contact_matrices_at_each_to_len = self.contact_matrices_at_each_to_len
    
    cdef int max_contact_matrices_used_at_single_node = np.max( np.concatenate([
        [contact_matrices_at_each_loc_len[i] for i in range(max_node_index+1)],
        [contact_matrices_at_each_to_len[i] for i in range(max_node_index+1)]
    ])) # Used to create _lambda

    # Transport
    cdef transporter* Ts = self.Ts # Going into commuterverses
    cdef int Ts_num = self.Ts_num
    cdef transporter* cTs = self.cTs # Going out from commuterverses
    cdef int cTs_num = self.cTs_num

    # Used for lambda calculation
    cdef int*** nodes_at_j = self.nodes_at_j
    cdef int** nodes_at_j_len = self.nodes_at_j_len

    # Used for tau calculation
    cdef int*** cnodes_into_k = self.cnodes_into_k
    cdef int** cnodes_into_k_len = self.cnodes_into_k_len
    
    # Transport profile
    cdef DTYPE_t transport_profile_integrated = self.transport_profile_integrated
    cdef DTYPE_t transport_profile_integrated_r = self.transport_profile_integrated_r
    cdef DTYPE_t transport_profile_m = self.transport_profile_m
    cdef DTYPE_t transport_profile_c = self.transport_profile_c
    cdef DTYPE_t transport_profile_c_r = self.transport_profile_c_r

    # Stochasticity
    cdef bint stochastic_simulation = self.stochastic_simulation
    total_Os_arr = np.zeros(model_dim, dtype=DTYPE)
    cdef DTYPE_t[:] total_Os = total_Os_arr # Used to see whether stochasticity should be turned on
    cdef DTYPE_t[:] stochastic_threshold_from_below = self.stochastic_threshold_from_below
    cdef DTYPE_t[:] stochastic_threshold_from_above = self.stochastic_threshold_from_above
    cdef bint* loc_j_is_stochastic
    cdef bint* to_k_is_stochastic

    # Contact matrix scaling
    cdef int contact_scaling_type = self.contact_scaling_type
    cdef int contact_scaling_types_linear = contact_scaling_types.linear
    cdef int contact_scaling_types_powerlaw = contact_scaling_types.powerlaw
    cdef int contact_scaling_types_exp = contact_scaling_types.exp
    cdef int contact_scaling_types_log = contact_scaling_types.log
    cdef DTYPE_t[:] contact_scaling_params = self.contact_scaling_params
    cdef DTYPE_t _f, _g, _cg
    
    location_r_area_arr = np.array([1.0/a if not np.isnan(a) else np.nan for a in self.location_area]) # Compute inverse of area
    cdef DTYPE_t[:] location_r_area = location_r_area_arr

    commuterverse_r_area_arr = np.array([1.0/a if not np.isnan(a) else np.nan for a in self.commuterverse_area])
    cdef DTYPE_t[:] commuterverse_r_area = commuterverse_r_area_arr

    cdef np.ndarray cmat_scaling_fg_arr = np.zeros( ( self.max_node_index+1, age_groups, age_groups) )
    cdef np.ndarray cmat_scaling_fg_cverse_arr = np.zeros( ( self.max_node_index+1, age_groups, age_groups) )
    cdef np.ndarray cmat_scaling_a_arr = np.zeros( (age_groups, age_groups) )
    cdef DTYPE_t[:,:,:] cmat_scaling_fg = cmat_scaling_fg_arr
    cdef DTYPE_t[:,:,:] cmat_scaling_fg_cverse = cmat_scaling_fg_cverse_arr
    cdef DTYPE_t[:,:] cmat_scaling_a = cmat_scaling_a_arr
     
    if random_seed == -1:
        random_seed = np.int64(np.round(time.time()))
    cdef mt19937 gen = mt19937(random_seed)
    cdef poisson_distribution[int] dist

    #if random_seed == -1:
    #    np.random.seed( np.int64(np.round(time.time())) )

    # Consts
    cdef int minutes_in_day = 1440
    save_node_mappings_path = 'node_mappings.pkl'
    save_cnode_mappings_path = 'cnode_mappings.pkl'
    save_ts_path = 'ts.npy'
    save_X_states_path = 'X_states.zarr'

    #### Simulation variables ##########################################

    cdef int cni, ni, si, Ti, cTi, i, j, ui, u, o, cmat_i, oi, X_index, loc_j, to_k, age_a, age_b, loc
    cdef bint to_k_is_active = False
    cdef DTYPE_t S, t1, t2, transport_profile, fro_N, cn_N, term, transport_profile_exponent, _N
    cdef node n, fro_n, to_n
    cdef cnode cn
    cdef model_term mt
    
    _lambdas_arr = np.zeros(( max_contact_matrices_used_at_single_node, self.age_groups, self.infection_classes_num ))
    cdef DTYPE_t[:, :, :] _lambdas = _lambdas_arr
    cdef DTYPE_t[:] _Is = self._Is
    cdef DTYPE_t[:] _Ns = self._Ns

    cdef np.ndarray dX_state_arr
    cdef DTYPE_t[:] dX_state

    #### Compute total population ######################################

    for i in range(nodes_num):
        n = self.nodes[i]
        for o in range(model_dim):
            total_N[n.age] += X_state[n.state_index + o]

    for i in range(cnodes_num):
        cn = self.cnodes[i]
        for o in range(model_dim):
            total_N[cn.age] += X_state[cn.state_index + o]

    #### Initialize stochasticity ######################################

    loc_j_is_stochastic = <bint *> malloc( (self.max_node_index+1) * sizeof(bint))
    to_k_is_stochastic = <bint *> malloc( (self.max_node_index+1) * sizeof(bint))

    if stochastic_simulation:

        for loc_j in range(max_node_index+1):
            n = nodes[i]

            for o in range(model_dim):
                total_Os[o] = 0

            for age_a in range(age_groups):
                for i in range(nodes_at_j_len[age_a][loc_j]):
                    n = nodes[nodes_at_j[age_a][loc_j][i]]
                    for o in range(model_dim):
                        total_Os[o] += X_state[n.state_index + o]

            loc_j_is_stochastic[loc_j] = False

            for o in range(model_dim):
                loc_j_is_stochastic[loc_j] = loc_j_is_stochastic[loc_j] or (total_Os[o] < stochastic_threshold_from_below[o])

        for to_k in range(max_node_index+1):
            n = nodes[i]

            for o in range(model_dim):
                total_Os[o] = 0

            for age_a in range(age_groups):
                for i in range(cnodes_into_k_len[age_a][to_k]):
                    cn = cnodes[cnodes_into_k[age_a][to_k][i]]
                    for o in range(model_dim):
                        total_Os[o] += X_state[n.state_index + o]

            to_k_is_stochastic[to_k] = False

            for o in range(model_dim):
                to_k_is_stochastic[to_k] = to_k_is_stochastic[to_k] or (total_Os[o] < stochastic_threshold_from_below[o])

    # Loop through nodes and cnodes to see if they should start out as stochastic or not

    #### Calcuate steps for the Forward-Euler integration ##############

    if type(_dts) == list:
        _dts= np.array(_dts)
    elif not type(_dts) == np.ndarray:
        _dts = np.array([_dts], dtype=DTYPE)
    dts = _dts

    t = t_start
    steps = 0
    while t <= t_end:
        t += dts[steps % len(dts)]
        steps += 1
    steps -= 1

    dX_state_arr = np.zeros( X_state.size )
    dX_state = dX_state_arr

    dts_num = len(dts)

    #### Set-up variables for storing the simulation history ###########

    cdef int save_i, X_states_saved_col_num
    X_states_saved = None
    ts_saved = None

    if steps_per_save != -1:

        # Compute the number of save states
        t = t_start
        save_states = 1
        for step_i in range(steps):
            t += dts[steps % len(dts)]
            if step_i % steps_per_save == 0:
                save_states += 1

        if only_save_nodes:
            X_states_saved_col_num = self.node_states_len
        else:
            X_states_saved_col_num = X_state.size

        # If save_path is not specified, then the states will be stored to numpy array
        if save_path == "":
            X_states_saved = np.zeros( ( save_states, X_states_saved_col_num) )

        # If save_path is specified, then the states will be stored on the harddrive directly, using a zarr array
        else: 
            X_states_saved = zarr.open('%s/%s' % (save_path, save_X_states_path), mode='w',
                                shape=( save_states, X_states_saved_col_num ),
                                chunks=(1 , X_states_saved_col_num), dtype=DTYPE)
            
        X_states_saved[0, :] = X_state_arr[:X_states_saved_col_num]
        ts_saved = np.zeros( X_states_saved.shape[0] )
        ts_saved[0] = t_start
        save_i = 1

    #### Event management ##############################################

    cdef list event_functions = self.event_functions
    cdef np.ndarray is_event_on_arr = np.zeros((steps, len(self.event_functions)), dtype=np.int8)
    cdef char[:,:] is_event_on = is_event_on_arr
    cdef int event_i

    is_a_event_on_arr = np.zeros(steps, dtype=np.int8)
    cdef char[:] is_a_event_on = is_a_event_on_arr

    # Align event times with simulation steps

    t = t_start
    for step_i in range(steps):
        dt = dts[step_i % dts_num]
        t += dt

        for event_i in range(len(self.event_times)):
            for et in self.event_times[event_i]:
                if self.event_repeat_times[event_i] == -1:
                    _t = t
                else:
                    _t = (t % self.event_repeat_times[event_i])

                if et <= _t and _t < et+dt:
                    is_event_on_arr[step_i][event_i] = 1
                    is_a_event_on[step_i] = 1

    #Ns0 = np.zeros( (max_node_index+1, age_groups) )
    #for loc in range(max_node_index+1):
    #    for age in range(age_groups):
    #        for i in range(nodes_at_j_len[age][loc]):
    #            n = nodes[nodes_at_j[age][loc][i]]

    #            for o in range(model_dim):
    #                Ns0[loc,age] += X_state[n.state_index + o]

    ####################################################################
    #### Simulation ####################################################
    ####################################################################

    t = t_start

    for step_i in range(steps):
        tday = t % minutes_in_day
        dt = dts[step_i % dts_num]
        r_dt = 1/dt

        # Reset dX_state to 0
        for i in prange(state_size, nogil=True):
            dX_state[i] = 0

        ################################################################
        #### Dynamics ##################################################
        ################################################################

        if measure_sim_times: t_contact = time.time()

        #### Contact matrix scaling ####################################

        for i in range(age_groups):
            for j in range(age_groups):
                cmat_scaling_a[i,j] = 0

        for loc in range(max_node_index+1):

            # Reset arrays
            for age in range(age_groups):
                Ns[loc,age] = 0
                cNs[loc,age] = 0
                for o in range(model_dim):
                    Os[loc,age,o] = 0
                    cOs[loc,age,o] = 0

            # Nodes
            for age in range(age_groups):
                for i in range(nodes_at_j_len[age][loc]):
                    n = nodes[nodes_at_j[age][loc][i]]

                    _N = 0
                    for o in range(model_dim):
                        Os[loc,age,o] += X_state[n.state_index + o]
                        _N += X_state[n.state_index + o]

                    if _N < 1e-1:
                        nodes[nodes_at_j[age][loc][i]].is_on = False
                    else:
                        nodes[nodes_at_j[age][loc][i]].is_on = True

                    Ns[loc,age] += _N

            # Cnodes
            for age in range(age_groups):
                for i in range(cnodes_into_k_len[age][loc]):
                    cn = cnodes[cnodes_into_k[age][loc][i]]

                    if cn.is_on:
                        _N = 0
                        for o in range(model_dim):
                            cOs[loc,age,o] += X_state[cn.state_index + o]
                            _N += X_state[cn.state_index + o]

                        cNs[loc,age] += _N

            # Compute scaling tensor

            for i in range(age_groups):
                for j in range(age_groups):

                    _g = Ns[loc, i]*Ns[loc, j] * location_r_area[loc]*location_r_area[loc]
                    _cg = cNs[loc, i]*cNs[loc, j] * commuterverse_r_area[loc]*commuterverse_r_area[loc]

                    if contact_scaling_type == contact_scaling_types_linear:
                        _g = contact_scaling_params[0] + contact_scaling_params[1]*_g
                        _cg = contact_scaling_params[0] + contact_scaling_params[1]*_cg
                    elif contact_scaling_type == contact_scaling_types_powerlaw:
                        _g = contact_scaling_params[0] + contact_scaling_params[1] * libc.math.pow(_g, contact_scaling_params[2])
                        _cg = contact_scaling_params[0] + contact_scaling_params[1] * libc.math.pow(_cg, contact_scaling_params[2])
                    elif contact_scaling_type == contact_scaling_types_exp:
                        _g = contact_scaling_params[0] * libc.math.pow(contact_scaling_params[1], _g)
                        _cg = contact_scaling_params[0] * libc.math.pow(contact_scaling_params[1], _cg)
                    elif contact_scaling_type == contact_scaling_types_log:
                        _g = libc.math.log(contact_scaling_params[0] + contact_scaling_params[1]*_g)
                        _cg = libc.math.log(contact_scaling_params[0] + contact_scaling_params[1]*_cg)

                    if False and _g!=_g:
                        print(1)
                        print(location_r_area[loc])
                        print(Ns[loc, i])
                        print(Ns[loc, j])
                        print(total_N[j])
                        print(_g)
                        print("---")

                    if False and _g == 0:
                        print(2)
                        print(location_r_area[loc])
                        print(Ns[loc, i], loc, i)
                        print(Ns[loc, j], loc, j)
                        print(_g)
                        print("---")
                    
                    if _g != 0 and _g==_g:
                        _f = libc.math.sqrt( total_N[i] * Ns[loc, j] / (Ns[loc, i] * total_N[j]) )
                        #_f = total_N[i] / Ns[loc, i]
                        cmat_scaling_fg[loc,i,j] = _f*_g
                        cmat_scaling_a[i,j] += Ns[loc, i]*cmat_scaling_fg[loc,i,j]
                        #cmat_scaling_a[i,j] += cmat_scaling_fg[loc,i,j]
                    else:
                        cmat_scaling_fg[loc,i,j] = 0

                    if _cg != 0 and _cg==_cg:
                        _f = libc.math.sqrt( total_N[i] * cNs[loc, j] / (cNs[loc, i] * total_N[j]) )
                        #_f = total_N[i] / cNs[loc, i]
                        cmat_scaling_fg_cverse[loc,i,j] = _f*_cg
                        cmat_scaling_a[i,j] += cNs[loc, i]*cmat_scaling_fg_cverse[loc,i,j]
                        #cmat_scaling_a[i,j] += cmat_scaling_fg_cverse[loc,i,j]
                    else:
                        cmat_scaling_fg_cverse[loc,i,j]=0

        for i in range(age_groups):
            for j in range(age_groups):
                cmat_scaling_a[i,j] = cmat_scaling_a[i,j] / total_N[i]

        # Normalize cmat_scaling_fg and cmat_scaling_fg_cverse using cmat_scaling_a

        for loc in range(max_node_index+1):
            for i in range(age_groups):
                for j in range(age_groups):
                    cmat_scaling_fg[loc,i,j] = cmat_scaling_fg[loc,i,j] / cmat_scaling_a[i,j]
                    cmat_scaling_fg_cverse[loc,i,j] = cmat_scaling_fg_cverse[loc,i,j] / cmat_scaling_a[i,j]

        if measure_sim_times: t_contact = time.time() - t_contact

        #### Node dynamics #############################################
        
        if measure_sim_times: t_node = time.time()

        for loc_j in range(max_node_index+1):

            #### Compute lambdas

            for ui in range(infection_classes_num):
                u = infection_classes_indices[ui]

                # Compute lambdas

                for cmat_i in range(contact_matrices_at_each_loc_len[loc_j]):
                    for age_a in range(age_groups):
                        _lambdas[cmat_i][age_a][ui] = 0
                        for age_b in range(age_groups):
                            if Ns[loc_j,age_b] > 1: # No infections can occur if there are fewer than one person at node
                                _lambdas[cmat_i][age_a][ui] += cmat_scaling_fg[loc_j][age_a][age_b]*contact_matrices[contact_matrices_at_each_loc[loc_j][cmat_i]][age_a][age_b] * Os[loc_j,age_b,u] / Ns[loc_j,age_b]

            #### Decide whether deterministic or stochastic

            for o in range(model_dim):
                total_Os[o] = 0
                for age in range(age_groups):
                    total_Os[o] += Os[loc_j,age,o]
            
            if stochastic_simulation:
                if loc_j_is_stochastic[loc_j]:
                    loc_j_is_stochastic[loc_j] = False
                    for o in range(model_dim):
                        loc_j_is_stochastic[loc_j] = loc_j_is_stochastic[loc_j] or (total_Os[o] < stochastic_threshold_from_below[o])
                else:
                    loc_j_is_stochastic[loc_j] = False
                    for o in range(model_dim):
                        loc_j_is_stochastic[loc_j] = loc_j_is_stochastic[loc_j] or (total_Os[o] < stochastic_threshold_from_above[o])
            
            #### Compute the derivatives at each node

            # Stochastic
            
            if stochastic_simulation and loc_j_is_stochastic[loc_j]:
                for age_a in range(age_groups):
                    for i in range(nodes_at_j_len[age_a][loc_j]): 
                        n = nodes[nodes_at_j[age_a][loc_j][i]]

                        if not n.is_on:
                            continue

                        si = n.state_index
                        S = X_state[si] # S is always located at the state index
                        
                        for j in range(model_linear_terms_len):
                            mt = model_linear_terms[j]   
                            if X_state[si+mt.oi_neg] > 0: # Only allow interaction if the class is positive
                                dist = poisson_distribution[int](dt*n.linear_coeffs[j]*X_state[si+mt.oi_coupling])
                                term = dist(gen) * r_dt
                                #term = scipy.stats.poisson.rvs(dt*n.linear_coeffs[j]*X_state[si+mt.oi_coupling]) * r_dt
                                dX_state[si+mt.oi_pos] += term
                                dX_state[si+mt.oi_neg] -= term

                        for j in range(model_infection_terms_len):
                            mt = model_infection_terms[j]
                            if X_state[si+mt.oi_neg] > 0: # Only allow interaction if the class is positive
                                cmat_i = n.contact_matrix_indices[mt.infection_index]
                                dist = poisson_distribution[int](dt*n.infection_coeffs[j]*_lambdas[cmat_i][age_a][mt.infection_index]*S)
                                term = dist(gen) * r_dt
                                #term = scipy.stats.poisson.rvs(dt*n.infection_coeffs[j]*_lambdas[cmat_i][age_a][mt.infection_index]*S) * r_dt

                                dX_state[si+mt.oi_pos] += term
                                dX_state[si+mt.oi_neg] -= term      
 
            # Deterministic
            else:
                for age_a in range(age_groups):
                    for i in range(nodes_at_j_len[age_a][loc_j]): 
                        n = nodes[nodes_at_j[age_a][loc_j][i]]

                        if not n.is_on:
                            continue

                        si = n.state_index
                        S = X_state[si] # S is always located at the state index
                        
                        for j in range(model_linear_terms_len):
                            mt = model_linear_terms[j]
                            if X_state[si+mt.oi_neg] > 0: # Only allow interaction if the class is positive
                                term = n.linear_coeffs[j] * X_state[si+mt.oi_coupling]
                                dX_state[si+mt.oi_pos] += term
                                dX_state[si+mt.oi_neg] -= term

                        for j in range(model_infection_terms_len):
                            mt = model_infection_terms[j]

                            if X_state[si+mt.oi_neg] > 0: # Only allow interaction if the class is positive
                                cmat_i = n.contact_matrix_indices[mt.infection_index]
                                term = n.infection_coeffs[j] * _lambdas[cmat_i][age_a][mt.infection_index] * S
                                dX_state[si+mt.oi_pos] += term
                                dX_state[si+mt.oi_neg] -= term

        if measure_sim_times: t_node = time.time() - t_node

        #### CNode dynamics ############################################

        if measure_sim_times: t_cnode = time.time()

        for to_k in range(max_node_index+1):

            # Check whether there are any active cnodes going into k

            to_k_is_active = False
            for age_a in range(age_groups):
                for i in range(cnodes_into_k_len[age_a][to_k]):
                    cn = cnodes[cnodes_into_k[age_a][to_k][i]]
                    to_k_is_active = cn.is_on or to_k_is_active
                    if to_k_is_active:
                        break
                if to_k_is_active:
                    break

            if not to_k_is_active:
                continue

            #### Compute lambdas

            for ui in range(infection_classes_num):
                u = infection_classes_indices[ui]

                # Compute lambdas

                for cmat_i in range(contact_matrices_at_each_to_len[to_k]):
                    for age_a in range(age_groups):
                        _lambdas[cmat_i][age_a][ui] = 0
                        for age_b in range(age_groups):
                            if cNs[to_k,age_b] > 1: # No infections can occur if there are fewer than one person at node
                                _lambdas[cmat_i][age_a][ui] += cmat_scaling_fg_cverse[to_k][age_a][age_b]*contact_matrices[contact_matrices_at_each_to[to_k][cmat_i]][age_a][age_b] * Os[to_k,age_b,u] / cNs[to_k,age_b]

            #### Decide whether deterministic or stochastic

            for o in range(model_dim):
                total_Os[o] = 0
                for age in range(age_groups):
                    total_Os[o] += Os[to_k,age,o]

            if stochastic_simulation:
                if to_k_is_stochastic[to_k]:
                    to_k_is_stochastic[to_k] = False
                    for o in range(model_dim):
                        to_k_is_stochastic[to_k] = to_k_is_stochastic[to_k] or (total_Os[o] < stochastic_threshold_from_below[o])
                else:
                    to_k_is_stochastic[to_k] = True
                    for o in range(model_dim):
                        to_k_is_stochastic[to_k] = to_k_is_stochastic[to_k] and (total_Os[o] > stochastic_threshold_from_above[o])

            # Stochastic
            if stochastic_simulation and to_k_is_stochastic[to_k]:
                for age_a in range(age_groups):
                    for i in range(cnodes_into_k_len[age_a][to_k]): 
                        cn = cnodes[cnodes_into_k[age_a][to_k][i]]

                        if not cn.is_on:
                            continue

                        si = cn.state_index
                        S = X_state[si] # S is always located at the state index

                        for j in range(model_linear_terms_len):
                            mt = model_linear_terms[j]
                            if X_state[si+mt.oi_neg] > 0: # Only allow interaction if the class is positive
                                dist = poisson_distribution[int](dt*cn.linear_coeffs[j]*X_state[si+mt.oi_coupling])
                                term = dist(gen) * r_dt
                                #term = scipy.stats.poisson.rvs(dt*cn.linear_coeffs[j]*X_state[si+mt.oi_coupling]) * r_dt
                                dX_state[si+mt.oi_pos] += term
                                dX_state[si+mt.oi_neg] -= term

                        for j in range(model_infection_terms_len):
                            mt = model_infection_terms[j]
                            if X_state[si+mt.oi_neg] > 0: # Only allow interaction if the class is positive
                                cmat_i = cn.contact_matrix_indices[mt.infection_index]
                                dist = poisson_distribution[int](dt*cn.infection_coeffs[j]*_lambdas[cmat_i][age_a][mt.infection_index]*S)
                                term = dist(gen) * r_dt
                                #term = scipy.stats.poisson.rvs(dt*cn.infection_coeffs[j]*_lambdas[cmat_i][age_a][mt.infection_index]*S) * r_dt
                                dX_state[si+mt.oi_pos] += term
                                dX_state[si+mt.oi_neg] -= term
            # Deterministic
            else:
                for age_a in range(age_groups):
                    for i in range(cnodes_into_k_len[age_a][to_k]): 
                        cn = cnodes[cnodes_into_k[age_a][to_k][i]]

                        if not cn.is_on:
                            continue

                        si = cn.state_index
                        S = X_state[si] # S is always located at the state index

                        for j in range(model_linear_terms_len):
                            mt = model_linear_terms[j]
                            if X_state[si+mt.oi_neg] > 0: # Only allow interaction if the class is positive
                                term = cn.linear_coeffs[j] * X_state[si+mt.oi_coupling]
                                dX_state[si+mt.oi_pos] += term
                                dX_state[si+mt.oi_neg] -= term

                        for j in range(model_infection_terms_len):
                            mt = model_infection_terms[j]
                            if X_state[si+mt.oi_neg] > 0: # Only allow interaction if the class is positive
                                cmat_i = cn.contact_matrix_indices[mt.infection_index]
                                term = cn.infection_coeffs[j]*_lambdas[cmat_i][age_a][mt.infection_index]*S
                                dX_state[si+mt.oi_pos] += term
                                dX_state[si+mt.oi_neg] -= term               
        
        if measure_sim_times: t_cnode = time.time() - t_cnode
        
        ################################################################
        #### Transport #################################################
        ################################################################

        if measure_sim_times: t_T = time.time()

        #### Node to CNode #############################################

        for Ti in range(Ts_num):

            t1 = Ts[Ti].t1
            t2 = Ts[Ti].t2

            if tday >= t1 and tday <= t2:
                fro_n = nodes[Ts[Ti].fro_node_index] # Origin node
                cn = cnodes[Ts[Ti].cnode_index] # Commuting node

                # Compute current population at origin node
                fro_N = 0
                for oi in range(model_dim):
                    fro_N += X_state[fro_n.state_index + oi]

                # If this commuting schedule is just starting, then
                # compute the number of people to move.
                if not Ts[Ti].is_on:
                    if Ts[Ti].use_percentage:
                        Ts[Ti].N0 = fro_N*Ts[Ti].move_percentage
                    else:
                        Ts[Ti].N0 = Ts[Ti].move_N
                    Ts[Ti].is_on = True
                    cnodes[Ts[Ti].cnode_index].is_on = True # Turn on commuter node. It will be turned off in the "CNode to Node" section

                # Compute the transport profile
                transport_profile_exponent = (tday - t1)*Ts[Ti].r_T_Delta_t - transport_profile_m
                transport_profile = libc.math.exp(- transport_profile_exponent * transport_profile_exponent * transport_profile_c_r) * transport_profile_integrated_r * Ts[Ti].r_T_Delta_t
                
                if fro_N <= 0:
                    continue
                
                si = fro_n.state_index
                for oi in range(model_dim):
                    if not Ts[Ti].moving_classes[oi]:
                        continue

                    # Compute the amount of people to move
                    term = Ts[Ti].N0 * transport_profile * (X_state[fro_n.state_index+oi] / fro_N)

                    # If the change will cause X_state[si+oi] to go negative,
                    # then adjust term so that X_state[si+oi] will be set 
                    # to 0.
                    if X_state[si+oi] + dt*(dX_state[si+oi] - term) < 0:
                        term = X_state[si+oi]*r_dt
                        dX_state[cn.state_index+oi] += term + dX_state[si+oi] # We shift the SIR dynamics that transpired in the node into the cnode
                        dX_state[si+oi] += -(term + dX_state[si+oi])
                    # Otherwise apply the transport as usual
                    else:
                        dX_state[si+oi] -= term
                        dX_state[cn.state_index+oi] += term

            else:
                Ts[Ti].is_on = False

        if measure_sim_times: t_T = time.time() - t_T
        
        #### CNode to Node #############################################

        if measure_sim_times: t_cT = time.time()
        
        for cTi in range(cTs_num):

            t1 = cTs[cTi].t1
            t2 = cTs[cTi].t2

            if tday >= t1 and tday <= t2:
                cn = cnodes[cTs[cTi].cnode_index] # Commuting node
                to_node = nodes[cTs[cTi].to_node_index] # Destination node

                # Compute current population at the commuter node
                cn_N = 0
                for oi in range(model_dim):#prange(model_dim, nogil=True):
                    cn_N += X_state[cn.state_index + oi]

                # If this commuting schedule is just starting, then
                # compute the number of people to move.
                if not cTs[cTi].is_on:
                    if cTs[cTi].use_percentage:
                        cTs[cTi].N0 = cn_N*cTs[cTi].move_percentage
                    else:
                        cTs[cTi].N0 = cTs[cTi].move_N
                    cTs[cTi].is_on = True

                # Compute the transport profile
                transport_profile_exponent = (tday - t1)*cTs[cTi].r_T_Delta_t - transport_profile_m
                transport_profile = libc.math.exp(- transport_profile_exponent * transport_profile_exponent * transport_profile_c_r) * transport_profile_integrated_r * cTs[cTi].r_T_Delta_t

                if cn_N <= 0:
                    continue

                si = cn.state_index
                for oi in range(model_dim):#prange(model_dim, nogil=True):
                    if not cTs[cTi].moving_classes[oi]:
                        continue

                    # If the commuting window is ending, force all to leave the commuterverse
                    if tday+dt >= t2:
                        term = X_state[si+oi]*r_dt
                        dX_state[to_node.state_index+oi] += term + dX_state[si+oi] # We shift the SIR dynamics that transpired in the cnode into the node
                        dX_state[si+oi] += - (term + dX_state[si+oi])
                    else:
                        # Compute the amount of people to move
                        term = cTs[cTi].N0 * transport_profile * (X_state[si+oi] / cn_N)

                        # If the change will cause X_state[si+oi] to go negative,
                        # then adjust term so that X_state[si+oi] will be set 
                        # to 0. 
                        if X_state[si+oi] + dt*(dX_state[si+oi] - term) < 0:
                            term = X_state[si+oi]*r_dt
                            dX_state[to_node.state_index+oi] += term + dX_state[si+oi] # We shift the SIR dynamics that transpired in the cnode into the node
                            dX_state[si+oi] += - (term + dX_state[si+oi])
                        # Otherwise apply the transport as usual
                        else:
                            dX_state[si+oi] -= term
                            dX_state[to_node.state_index+oi] += term
            else:
                cTs[cTi].is_on = False
                cn = cnodes[cTs[cTi].cnode_index]
                cn.is_on = False # Turn off commuter node
                
        if measure_sim_times: t_cT = time.time() - t_cT

        ################################################################
        #### Forward-Euler #############################################
        ################################################################

        if measure_sim_times: t_ff = time.time()
        
        for j in prange(X_state_size, nogil=True):
            X_state[j] += dX_state[j]*dt

        t += dt

        if measure_sim_times: t_ff = t_ff - time.time()
        
        if steps_per_print != -1 and step_i % steps_per_print==0:
            print("Step %s out of %s" % (step_i, steps))

            #for loc in range(max_node_index+1):
            #    for age in range(age_groups):
            #        if abs(Ns0[loc, age] - Ns[loc, age]) > 0.1:
            #            print(loc, age, Ns0[loc, age], Ns[loc, age])

        if measure_sim_times: t_rest = time.time()
        
        #### Store state

        if steps_per_save != -1:

            if (step_i+1) % steps_per_save == 0:
                X_states_saved[save_i,:] = X_state[:X_states_saved_col_num]
                ts_saved[save_i] = t
                save_i += 1

        #### Call event function

        if is_a_event_on[step_i]:
            event_i = 0
            for event_function in event_functions:
                if is_event_on[step_i][event_i]:
                    if steps_per_save == -1:
                        event_function(self, step_i, t, dt, X_state, dX_state)
                    else:
                        event_function(self, step_i, t, dt, X_state, dX_state, X_states_saved, ts_saved, save_i)
                event_i += 1

        #### Call Cython event function

        #if cevent_steps[step_i]:
        #    cevent_function(self, step_i, t, dt, X_state, dX_state)

        if measure_sim_times: t_rest = time.time() - t_rest

        if measure_sim_times:
            print("""
            step = %s, time = %s
            t_contact = %s
            t_node = %s
            t_cnode = %s
            t_T = %s
            t_cT = %s
            t_ff = %s
            t_rest = %s
            t_total = %s
            --------------
            """ % (step_i, t, t_contact, t_node, t_cnode, t_T, t_cT, t_ff, t_rest, t_contact+t_node+t_cnode+t_T+t_cT+t_ff+t_rest))

    free(loc_j_is_stochastic)
    free(to_k_is_stochastic)

    if steps_per_save != -1:

        node_mappings = self.node_mappings.copy()
        cnode_mappings = self.cnode_mappings.copy()

        sim_data = (node_mappings, cnode_mappings, ts_saved, X_states_saved)

        if save_path != '':
            with open("%s/%s" % (save_path, save_node_mappings_path),"wb") as f:
                pickle.dump(node_mappings, f)
            with open("%s/%s" % (save_path, save_cnode_mappings_path),"wb") as f:
                pickle.dump(cnode_mappings, f)
            np.save("%s/%s" % (save_path, save_ts_path), ts_saved)

        return sim_data


import numpy                                                                                                    
import os, re, sys
#from distutils.core import setup                                                                                      
#from distutils.extension import Extension                                                                             
from setuptools import setup, Extension
from Cython.Build import cythonize                                                                                    
import Cython.Compiler.Options                                                                                        
Cython.Compiler.Options.annotate = True                                                                               
                                                                                                                      
                                                                                                                      
if 'darwin'==(sys.platform).lower():                                                                                  
    extension1 = Extension('pyrossgeo/*', ['pyrossgeo/*.pyx'],                                                           
        include_dirs=[numpy.get_include()],                                                                           
        extra_compile_args=['-mmacosx-version-min=10.9'],                                                             
        extra_link_args=['-mmacosx-version-min=10.9'],                                                                
    )                                                                                                                 
else:                                                                                                                 
    extension1 = Extension('pyrossgeo/*', ['pyrossgeo/*.pyx'],                                                           
        include_dirs=[numpy.get_include()],                                                                           
    )                                                                                                                 
                                                                                                                      
if 'darwin'==(sys.platform).lower():                                                                                  
    extension2 = Extension('pyrossgeo.mft.*', ['pyrossgeo/mft/*.pyx'],                                                       
        include_dirs=[numpy.get_include()],                                                                           
        extra_compile_args=['-mmacosx-version-min=10.9'],                                                             
        extra_link_args=['-mmacosx-version-min=10.9'],                                                                
    )                                                                                                                 
else:                                                                                                                 
    extension2 = Extension('pyrossgeo.mft.*', ['pyrossgeo/mft/*.pyx'],
        include_dirs=[numpy.get_include()],                                                                           
    )                                                                                                                 
                                                                                                                      
                                                                                                                      
with open("README.md", "r") as fh:
    long_description = fh.read()


cwd = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(cwd, 'pyrossgeo', '__init__.py')) as fp:
    for line in fp:
        m = re.search(r'^\s*__version__\s*=\s*([\'"])([^\'"]+)\1\s*$', line)
        if m:
            version = m.group(2)
            break
    else:
        raise RuntimeError('Unable to find own __version__ string')


setup(                                                                                                                
    name='pyrossgeo',                                                                                                 
    version=version,                                                                                                  
    url='https://github.com/lukastk/PyRossGeo.git',                                                                   
    project_urls={
            "Documentation": "https://pyrossgeo.readthedocs.io",
            "Source": "https://github.com/lukastk/PyRossGeo.git",
            },
    author='The PyRossGeo team',                                                                                      
    license='MIT',                                                                                                    
    description='PyRossGeo is a python library for spatially resolved numerical simulation of infectious diseases',                             
    long_description=long_description,
    long_description_content_type='text/markdown',
    platforms='works on all platforms (such as LINUX, macOS, and Microsoft Windows)',                                 
    ext_modules=cythonize([extension1, extension2],                                                                   
        compiler_directives={"language_level": 3},                                              
        ),                                                                                                            
    libraries=[],                                                                                                     
    packages=['pyrossgeo', 'pyrossgeo/mft'],                                                                                           
    package_data={
                'pyrossgeo': ['*pyx', '*.pxd'],                                                                            
                'pyrossgeo/mft': ['*pyx', '*.pxd']},                                                                            
    install_requires=['cython','numpy','scipy','matplotlib',
                     'pandas','zarr','geopandas'],
    include_package_data=True,
    classifiers=[
                'License :: OSI Approved :: MIT License',
                'Programming Language :: Python :: 3',
                'Programming Language :: Python :: 3.7',
                'Topic :: Scientific/Engineering',
                'Topic :: Scientific/Engineering :: Mathematics',
                'Intended Audience :: Science/Research',
                'Intended Audience :: Education',
                ],
)                                                                                                                     
